#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include <error.h>
#include <unistd.h>
#include <mpi.h>
#include <unistd.h>
#include "defs.h"

#define MIN_COMMUNITY_SIZE 10

/* one community size may be more than MAX_COMMUNITY_SIZE, if at the last step
   of generation, count of remaining vertices < MIN_COMMUNITY_SIZE */
#define MAX_COMMUNITY_SIZE 100000

/* must be < MIN_COMMUNITY_SIZE */
#define MIN_VERTEX_DEGREE 2

/* must be < MAX_COMMUNITY_SIZE */
#define MAX_VERTEX_DEGREE 25000

#define COMMUNITY_TAU 2.6
#define DEGREE_TAU 1.99
#define MU_DEFAULT 0.3
using namespace std;
void power_law(vertex_id_t smin, vertex_id_t smax, double tau, vector<double> &vals)
{
    vals.clear();
    double sum = 0;
    for (vertex_id_t i = smin; i <= smax; i++) {
        sum += pow((1.0 / i), tau);
    }
    
    double p = 0;
    for (vertex_id_t i = smin; i <= smax; i++) {
        p += pow((1.0 / i), tau);
        vals.push_back(p / sum);
    }
}

bool cmp(const vertex_id_t &x, const vertex_id_t &y)
{
    return x > y;
}


extern LFR_params LFR_p;
void gen_LFR_graph_MPI(graph_t* g)
{
    uint32_t local_TotVertices, TotVertices;
	uint32_t  local_totClusters;
	
	uint32_t *startVertex, *endVertex;
	uint32_t local_numEdges;
	weight_t* weights;
    weight_t MinWeight,MaxWeight;
    vertex_id_t u, v; 
    edge_id_t offset;
	uint32_t edgeNum;

    double mu = MU_DEFAULT;
	uint32_t i;
    int seed;
    int size, rank, lgsize;
    double community_tau = COMMUNITY_TAU, degree_tau = DEGREE_TAU;
    vertex_id_t smin = MIN_COMMUNITY_SIZE, smax = MAX_COMMUNITY_SIZE;
    vertex_id_t dmin = MIN_VERTEX_DEGREE, dmax = MAX_VERTEX_DEGREE;
    if(LFR_p.mu > -1) mu = LFR_p.mu;
    if(LFR_p.ctau > 0) community_tau = LFR_p.ctau;
    if(LFR_p.dtau > 0) degree_tau = LFR_p.ctau;
    if(LFR_p.mins > 0) smin = LFR_p.mins;
    if(LFR_p.maxs > 0) smax= LFR_p.maxs;
    if(LFR_p.maxd > 0) dmax= LFR_p.maxd;
    if(LFR_p.mind > 0) dmin = LFR_p.mind;

    g->directed = false;
    g->min_weight = 0;
    g->max_weight = 1;
    g->filename[0] = '\0';
    g->n = (vertex_id_t)1 << g->scale;
    if (strlen(g->filename) == 0) sprintf(g->filename, "LFR-%d", g->scale);

    /*----------------------------------------------*/
	/*		initialize 		*/
	/*----------------------------------------------*/
    seed = 1234;   
    srand48(time(NULL));
   
    MinWeight = g->min_weight;
    MaxWeight = g->max_weight;
    TotVertices = g->n;
    size = g->nproc;
    rank = g->rank;
    g->local_n = get_local_n(g);
    local_TotVertices = g->local_n;
    for (lgsize = 0; lgsize < size; ++lgsize) {
        if ((1 << lgsize) == size) break;
    }
    
    MPI_Datatype MPI_VERTEX_ID_T;                                        
    MPI_Type_contiguous(sizeof(vertex_id_t), MPI_BYTE, &MPI_VERTEX_ID_T);
    MPI_Type_commit(&MPI_VERTEX_ID_T);                                   

   
    //for (int i = 0; i < degree.size(); i++) {
    //    cout << i << " " << degree[i] << endl;
    //}
 
    /*----------------------------------------------*/
    /*             generate clusters                */
    /*----------------------------------------------*/
    
    vector<vertex_id_t> clusterSizes;
    vector<double> vals;
    vertex_id_t max_cluster = 0;
//    if (rank == 0) cout << "generate cluster sizes..." << endl;
    
    srand48(time(NULL));
    
    power_law(smin, smax, community_tau, vals);
    vertex_id_t cur_n = 0;
    while (cur_n < local_TotVertices) {
        vertex_id_t cur_size = lower_bound(vals.begin(), vals.end(), drand48()) - vals.begin() + smin;
        if ((long long)local_TotVertices - (long long)(cur_n + cur_size) < smin) {
            cur_size = local_TotVertices - cur_n;
        }
        assert(cur_size <= smax); 
        if (cur_size > max_cluster) max_cluster = cur_size;
        clusterSizes.push_back(cur_size);
        cur_n += cur_size;
    }
    
    local_totClusters = clusterSizes.size();
    /*----------------------------------------------*/
    /*      generate degrees of all vertex          */
    /*----------------------------------------------*/
//    vertex_id_t max_degree = 0;
//    if(rank == 0) cout << "generate degrees of all vertex..." << endl;
     
    vector<vertex_id_t> degree;
    power_law(dmin, min(dmax,max_cluster), degree_tau, vals);
    for (vertex_id_t i = 0; i < local_TotVertices; i++) {
        vertex_id_t cur_degree = lower_bound(vals.begin(), vals.end(), drand48()) - vals.begin() + dmin;
//        if (cur_degree > max_degree) max_degree = cur_degree;
        degree.push_back(cur_degree);
    }
    
    sort(degree.begin(), degree.end(), cmp);
 

    /*--------------------------------------------------*/
    /*  calc corresponds between vertices and clusters  */
    /*--------------------------------------------------*/
    
//    if(rank == 0) cout << "calc corresponds between vertices and clusters..." << endl;
    
    vector< vector<vertex_id_t> > vertices_in_cluster(local_totClusters);
    vector<vertex_id_t> cluster_num(local_TotVertices);
    vector<vertex_id_t> cur_cluster_size(local_totClusters, 0);
    vector<vertex_id_t> norm_clusters;
    for (vertex_id_t i = 0; i < local_totClusters; i++) {
        norm_clusters.push_back(i);
    }
    
    vertex_id_t norm_cluster_cnt = local_totClusters;
    for (vertex_id_t i = 0; i < local_TotVertices; i++) {
        vertex_id_t deg = degree[i];
        while (true) {
            vertex_id_t id = drand48() * norm_cluster_cnt;
            vertex_id_t num = norm_clusters[id];
            if (clusterSizes[num] > deg) {
                cluster_num[i] = num;
                vertices_in_cluster[num].push_back(i);
                cur_cluster_size[num]++;
                if (cur_cluster_size[num] == clusterSizes[num]) {
                    norm_clusters.erase(norm_clusters.begin() + id);
                    norm_cluster_cnt--;
                }
                
                break;
            }
        }
    }

    /*------------------------------------------------------*/
    /*                   generate edges                     */
    /*------------------------------------------------------*/
    
//    if(rank == 0) cout << "generate edges..." << endl;
    
    edgeNum = 0;
    vector<vertex_id_t> startV;
    vector<vertex_id_t> endV;
    
    for (vertex_id_t i = 0; i < local_TotVertices; i++) {
        vertex_id_t deg = degree[i];
        vertex_id_t outside_cnt = deg * mu;
        vertex_id_t inside_cnt = deg - outside_cnt;
        vertex_id_t num = cluster_num[i];
        vertex_id_t size_c = clusterSizes[num];
        for (vertex_id_t j = 0; j < inside_cnt; j++) {
            vertex_id_t k = size_c * drand48();
            vertex_id_t v = vertices_in_cluster[num][k];
            startV.push_back(VERTEX_TO_GLOBAL(i, TotVertices, size, rank));
            endV.push_back(VERTEX_TO_GLOBAL(v, TotVertices, size, rank));
            edgeNum++;
        }
        
        for (vertex_id_t j = 0; j < outside_cnt; j++) {
            vertex_id_t v;
            while (true) {
                v = TotVertices * drand48();
                if (v >= local_TotVertices) {
                    break;
                }
                if (cluster_num[v] != num) {
                    v = VERTEX_TO_GLOBAL(v, TotVertices, size, rank);
                    break;
                }
            }
            
            startV.push_back(VERTEX_TO_GLOBAL(i, TotVertices, size, rank));
            endV.push_back(v);
            edgeNum++;
        }
    }
//   cout<<startV.size()*2<<endl;
//return;
    local_numEdges = edgeNum;
 
    MPI_Barrier(MPI_COMM_WORLD); 
//    if(rank == 0) cout << "end of generating edges..." << endl;

startVertex = (uint32_t *) malloc(local_numEdges*sizeof(uint32_t));
endVertex = (uint32_t *) malloc(local_numEdges*sizeof(uint32_t));
for (i=0; i<local_numEdges; i++) {
    startVertex[i] = startV[i];
    endVertex[i] = endV[i];
}


//if(rank == 0) cout << "recive data..." << endl;
    /*----------------------------------------------*/
    /*              generate edge weights           */
    /*----------------------------------------------*/

    weights = (weight_t *) malloc(local_numEdges*sizeof(weight_t));

    for (i=0; i<local_numEdges; i++) {
        weights[i] = MinWeight + (weight_t) (MaxWeight - MinWeight) * drand48();
    }

    g->weights = weights;
	g->n = TotVertices;
    /*----------------------------------------------*/
    /*              calc data to send               */
    /*----------------------------------------------*/
    weight_t *send_weight = (weight_t*) malloc (2*local_numEdges * sizeof(weight_t));
    assert(send_weight != NULL);
    weight_t *recv_weight;
    vertex_id_t* send_edges = (vertex_id_t*) malloc (4 * local_numEdges * sizeof(vertex_id_t));
    assert(send_edges != NULL);
    vertex_id_t* recv_edges;
    int* recv_counts = (int*) calloc (size, sizeof(int));
    assert(recv_counts != NULL);
    int* send_counts = (int*) calloc (size, sizeof(int));
    assert(send_counts != NULL);    
    edge_id_t* send_offsets_edge = (edge_id_t*) calloc (size, sizeof(edge_id_t));
    assert(send_offsets_edge != NULL);
    edge_id_t* send_offsets_weight = (edge_id_t*) calloc (size, sizeof(edge_id_t));
    assert(send_offsets_weight != NULL);
    edge_id_t* recv_offsets_weight = (edge_id_t*) calloc (size, sizeof(edge_id_t));
    assert(recv_offsets_weight != NULL);
    edge_id_t* recv_offsets_edge = (edge_id_t*) calloc (size, sizeof(edge_id_t));
    assert(recv_offsets_edge != NULL);
	
	int proc_id;
    for (i=0; i<local_numEdges; i++) {
        proc_id = VERTEX_OWNER(startVertex[i], TotVertices, size);
        send_counts[proc_id]++;
        proc_id = VERTEX_OWNER(endVertex[i], TotVertices, size);
        send_counts[proc_id]++;
    }
    
    /*  calc offsets  */
    for (int i=1; i<size; i++) {
        send_offsets_edge[i] = send_offsets_edge[i-1] + 2 * send_counts[i-1];
        send_offsets_weight[i] = send_offsets_weight[i-1] + send_counts[i-1];
    }   
    for (int i=0; i<size; i++) {
        send_counts[i] = 0;
    }
    /* copy edges and weight to send_data */
    for (edge_id_t i=0; i<local_numEdges; i++) {
        proc_id = VERTEX_OWNER(startVertex[i], TotVertices, size);
        offset = send_offsets_edge[proc_id] + 2 * send_counts[proc_id];
        send_edges[offset + 0] = startVertex[i];
        send_edges[offset + 1] = endVertex[i];
        offset = send_offsets_weight[proc_id] + send_counts[proc_id];
        send_weight[offset] = weights[i];
        send_counts[proc_id]++;
         
        proc_id = VERTEX_OWNER(endVertex[i], TotVertices, size);
        offset = send_offsets_edge[proc_id] + 2*send_counts[proc_id];
        send_edges[offset + 0] = endVertex[i];
        send_edges[offset + 1] = startVertex[i];
        offset = send_offsets_weight[proc_id] + send_counts[proc_id];
        send_weight[offset] = weights[i];
        send_counts[proc_id]++; 
    }
    free(startVertex);
    free(endVertex);
    free(weights);
    
    MPI_Request request[size];
    MPI_Status status[size];

    MPI_Alltoall(send_counts, 1, MPI_INT, recv_counts, 1, MPI_INT, MPI_COMM_WORLD);
    
    for (int i=1; i<size; i++) {
        recv_offsets_weight[i] = recv_offsets_weight[i-1] + recv_counts[i-1];
    }
    edge_id_t counts = 0;
    for(int i=0; i<size; i++) {
        counts += recv_counts[i];
    }
    recv_weight = (weight_t*) malloc (counts * sizeof(weight_t));
    assert(recv_weight != NULL);
    
    MPI_Allreduce(&counts, &g->m, 1, MPI::UNSIGNED_LONG_LONG, MPI_SUM, MPI_COMM_WORLD); 

    /* send weights to each process */
    for (int i = 0; i < size; i++) {
        MPI_Irecv(&recv_weight[recv_offsets_weight[i]], recv_counts[i], MPI_DOUBLE, i, rank, MPI_COMM_WORLD, &request[i]);
    }
    for (int i = 0; i < size; i++) {
        MPI_Send(&send_weight[send_offsets_weight[i]], send_counts[i], MPI_DOUBLE, i, i, MPI_COMM_WORLD);
    }
    MPI_Waitall(size, request, status);

    free(send_weight);
    free(send_offsets_weight);
    free(recv_offsets_weight);

    /* calc offsets and number of elements for the next MPI_Send */
    for (int i=0; i<size; i++) {
        recv_counts[i] = 2*recv_counts[i];
        send_counts[i] = 2*send_counts[i];
    }
    for (int i=1; i<size; i++) {
        recv_offsets_edge[i] = recv_offsets_edge[i-1] + recv_counts[i-1];
    }
    recv_edges = (vertex_id_t*) malloc (2*counts * sizeof(vertex_id_t));
    assert(recv_edges != NULL);

    /* send edges to each process */
    for (int i = 0; i < size; i++) {
        MPI_Irecv(&recv_edges[recv_offsets_edge[i]], recv_counts[i], MPI_DOUBLE, i, rank, MPI_COMM_WORLD, &request[i]);
    }
    for (int i = 0; i < size; i++) {
        MPI_Send(&send_edges[send_offsets_edge[i]], send_counts[i], MPI_VERTEX_ID_T, i, i, MPI_COMM_WORLD);
    }
    MPI_Waitall(size, request, status);

    g->local_m = counts;
    vertex_id_t * degree_r = (vertex_id_t *) calloc(g->local_n, sizeof(vertex_id_t));
    assert(degree_r != NULL);
    for (edge_id_t i=0; i<2*g->local_m; i=i+2) {
        degree_r[VERTEX_LOCAL(recv_edges[i], TotVertices, size, rank)]++;
    }
    
    g->endV = (vertex_id_t *) calloc(g->local_m, sizeof(vertex_id_t));   
    assert(g->endV != NULL);                                             
    g->rowsIndices = (edge_id_t *) malloc((g->local_n+1)*sizeof(edge_id_t));
    assert(g->rowsIndices != NULL);                                      
                                                                     
    g->weights = (weight_t *) malloc(g->local_m * sizeof(weight_t));     
    assert(g->weights != NULL);                                          
    g->rowsIndices[0] = 0;                                               
    for (vertex_id_t i=1; i<=g->local_n; i++) {                          
        g->rowsIndices[i] = g->rowsIndices[i-1] + degree_r[i-1];           
    }                                                                    
                                                                     
    for (edge_id_t i=0; i<2*g->local_m; i=i+2) {                         
        u = VERTEX_LOCAL(recv_edges[i+0], TotVertices, size, rank);
        v = recv_edges[i+1];
        offset = degree_r[u]--;
        g->endV[g->rowsIndices[u]+offset-1] = v;
        g->weights[g->rowsIndices[u]+offset-1] = recv_weight[i/2];
    }    
    free(recv_edges);                                                    
    free(degree_r);                                                        
    free(recv_weight);                                                   

}
