#include "defs.h"
#include "louvain_mpi.h"
using namespace std;

double modularity_reference = -1;


void get_out_component_id(vertex_id_t n, set<vertex_id_t> &out_components, map<vertex_id_t, vertex_id_t> &out_component_id, const vector<vertex_id_t> &component_id)
{
	int size, rank;
	MPI_Comm_size(MPI_COMM_WORLD, &size);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	int *send_cnt = new int[size];
    int *recv_cnt = new int[size];
    memset(send_cnt, 0, sizeof(int) * size);
    for (set<vertex_id_t>::iterator i = out_components.begin(); i != out_components.end(); i++) {
        send_cnt[VERTEX_OWNER(*i, n, size)]++;
    }
	
    MPI_Alltoall(send_cnt, 1, MPI_INT, recv_cnt, 1, MPI_INT, MPI_COMM_WORLD);
	
    int **recv_buf = new int *[size];
    for (int i = 0; i < size; i++) {
        recv_buf[i] = new int[recv_cnt[i]];
    }
	
    MPI_Request *requests = new MPI_Request[size];
    for (int i = 0; i < size; i++) {
        MPI_Irecv(recv_buf[i], recv_cnt[i], MPI_INT, i, 0, MPI_COMM_WORLD, &requests[i]);
    }
	
    MPI_Barrier(MPI_COMM_WORLD);
	
    int **send_buf = new int *[size];
    for (int i = 0; i < size; i++) {
        send_buf[i] = new int[send_cnt[i]];
    }
    
    memset(send_cnt, 0, sizeof(int) * size);
    for (set<vertex_id_t>::iterator i = out_components.begin(); i != out_components.end(); i++) {
        int owner = VERTEX_OWNER(*i, n, size);
        send_buf[owner][send_cnt[owner]++] = *i;
    }
    
    for (int i = 0; i < size; i++) {
        MPI_Send(send_buf[i], send_cnt[i], MPI_INT, i, 0, MPI_COMM_WORLD);
    }
    
    MPI_Waitall(size, requests, MPI_STATUSES_IGNORE);
	
    int **send_buf2 = new int *[size];
    for (int i = 0; i < size; i++) {
        send_buf2[i] = new int[recv_cnt[i]];
    }
    
    int **recv_buf2 = new int *[size];
    for (int i = 0; i < size; i++) {
        recv_buf2[i] = new int[send_cnt[i]];
    }
    
    for (int i = 0; i < size; i++) {
        MPI_Irecv(recv_buf2[i], send_cnt[i], MPI_INT, i, MPI_ANY_TAG, MPI_COMM_WORLD, &requests[i]);
    }
	
    MPI_Barrier(MPI_COMM_WORLD);
	
	memset(send_cnt, 0, sizeof(int) * size);
    for (int i = 0; i < size; i++) {
        for (int j = 0; j < recv_cnt[i]; j++) {
            send_buf2[i][send_cnt[i]++] = component_id[VERTEX_LOCAL(recv_buf[i][j], n, size, rank)];
        }
    }
    
    for (int i = 0; i < size; i++) {
        MPI_Send(send_buf2[i], send_cnt[i], MPI_INT, i, 0, MPI_COMM_WORLD);
    }
    
    MPI_Waitall(size, requests, MPI_STATUSES_IGNORE);
	
    memset(send_cnt, 0, sizeof(int) * size);
    for (set<vertex_id_t>::iterator i = out_components.begin(); i != out_components.end(); i++) {
        int owner = VERTEX_OWNER(*i, n, size);
        vertex_id_t id = recv_buf2[owner][send_cnt[owner]++];
        out_component_id[*i] = id;
    }
	
	delete[] requests;
    delete[] send_cnt;
    delete[] recv_cnt;
    
	free_buf(send_buf, size);
    free_buf(send_buf2, size);
    free_buf(recv_buf, size);
    free_buf(recv_buf2, size);
}

double modularity(graph_t &graph, const vector<vertex_id_t> &component_id)
{
	vertex_id_t n = graph.n;
	vertex_id_t local_n = graph.local_n;
	int size, rank;
    MPI_Comm_size(MPI_COMM_WORLD, &size);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    map<vertex_id_t, vertex_id_t> out_component_id;
    set<vertex_id_t> out_components;
    sigma_map_t local_sigma;
    sigma_map_t sigma;
    weight_t local_m = 0;
    for (vertex_id_t i = 0; i < local_n; i++) {
        for (edge_id_t e = graph.rowsIndices[i]; e < graph.rowsIndices[i + 1]; e++) {
            out_components.insert(graph.endV[e]);
        }
    }
	
	get_out_component_id(n, out_components, out_component_id, component_id);
	
    for (vertex_id_t i = 0; i < local_n; i++) {
        for (edge_id_t e = graph.rowsIndices[i]; e < graph.rowsIndices[i + 1]; e++) {
            weight_t weight = graph.weights[e];
            local_m += weight;
            vertex_id_t cu = component_id[i], cv = out_component_id[graph.endV[e]];
            if (cu == cv) {
                local_sigma[cu].in += weight;
                local_sigma[cu].tot += weight;
            } else {
                local_sigma[cu].tot += weight;
                local_sigma[cv].tot += weight;
            }
        }
    }
	
    weight_t m;
    MPI_Allreduce(&local_m, &m, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
	
	int *send_cnt = new int[size];
    int *recv_cnt = new int[size];
    memset(send_cnt, 0, sizeof(int) * size);
    for (sigma_map_t::iterator i = local_sigma.begin(); i != local_sigma.end(); i++) {
        vertex_id_t cu = i->first;
        send_cnt[VERTEX_OWNER(cu, n, size)] += 2;
    }
	
    MPI_Alltoall(send_cnt, 1, MPI_INT, recv_cnt, 1, MPI_INT, MPI_COMM_WORLD);
    
	weight_t **send_buf = new weight_t *[size];
	for (int i = 0; i < size; i++) {
		send_buf[i] = new weight_t[send_cnt[i]];
	}
	
	weight_t **recv_buf = new weight_t *[size];
	for (int i = 0; i < size; i++) {
		recv_buf[i] = new weight_t[recv_cnt[i]];
	}
	
	int **send_buf2 = new int *[size];
	for (int i = 0; i < size; i++) {
		send_buf2[i] = new int[send_cnt[i]];
	}
	
	int **recv_buf2 = new int *[size];
	for (int i = 0; i < size; i++) {
		recv_buf2[i] = new int[recv_cnt[i]];
	}
	
	MPI_Request *requests = new MPI_Request[size];
	for (int i = 0; i < size; i++) {
		MPI_Irecv(recv_buf[i], recv_cnt[i], MPI_DOUBLE, i, 0, MPI_COMM_WORLD, &requests[i]);
	}
	
	MPI_Barrier(MPI_COMM_WORLD);
	
	memset(send_cnt, 0, sizeof(int) * size);
    for (sigma_map_t::iterator i = local_sigma.begin(); i != local_sigma.end(); i++) {
        vertex_id_t cu = i->first;
		int owner = VERTEX_OWNER(cu, n, size);
		send_buf[owner][send_cnt[owner]] = i->second.in;
		send_buf2[owner][send_cnt[owner]++] = 2 * cu;
		send_buf[owner][send_cnt[owner]] = i->second.tot;
		send_buf2[owner][send_cnt[owner]++] = 2 * cu + 1;
    }
	
	for (int i = 0; i < size; i++) {
		MPI_Send(send_buf[i], send_cnt[i], MPI_DOUBLE, i, 0, MPI_COMM_WORLD);
	}
	
	MPI_Waitall(size, requests, MPI_STATUSES_IGNORE);
	
	for (int i = 0; i < size; i++) {
        MPI_Irecv(recv_buf2[i], recv_cnt[i], MPI_INT, i, 0, MPI_COMM_WORLD, &requests[i]);
    }
	
	MPI_Barrier(MPI_COMM_WORLD);
	
	for (int i = 0; i < size; i++) {
		MPI_Send(send_buf2[i], send_cnt[i], MPI_INT, i, 0, MPI_COMM_WORLD);
	}
	
	MPI_Waitall(size, requests, MPI_STATUSES_IGNORE);
	
    for (int i = 0; i < size; i++) {
		for (int j = 0; j < recv_cnt[i]; j++) {
        	weight_t tmp = recv_buf[i][j];
        	int tag = recv_buf2[i][j];
        	vertex_id_t cu = tag / 2;
        	if (tag % 2) {
            	sigma[cu].tot += tmp;
        	} else {
            	sigma[cu].in += tmp;
        	}
		}
    }
	
    weight_t local_q = 0;
    for (sigma_map_t::iterator i = sigma.begin(); i != sigma.end(); i++) {
        local_q += i->second.in / m - pow(i->second.tot / m, 2);
    }
	
    weight_t q;
    MPI_Allreduce(&local_q, &q, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
	
    delete[] requests;
    delete[] send_cnt;
    delete[] recv_cnt;
	
	free_buf(send_buf, size);
	free_buf(send_buf2, size);
	
	free_buf(recv_buf, size);
	free_buf(recv_buf2, size);
	
    return q;
}

/* ratio == modularity / modularity_reference */
double f_exp(double ratio) 
{
    double c_point = 0.70;
    double c = 150;
    double alpha = log( c ) / (0.9 - c_point);
    double a = 0.9 / (exp ( 0.9 * alpha )); 
    
    return a * exp (alpha * ratio); 
}

/* rating function
   ratio == modularity / modularity_reference */
double rating(double time, double ratio) 
{
    double f;
    if (ratio >= 1) {
       f = (ratio - 1)*10 + 1;
    }   
    else if ((ratio > 0.9) && (ratio < 1)) {
        f = ratio;
    }   
    else {
      f = f_exp(ratio);
    }   
    double rating = 10000 * f / time;
    return rating;
}

void usage(int argc, char **argv)
{
    printf("Usage:\n");
    printf("  -s <scale> -- number of vertices is 2^<scale> (default scale: 23)\n");
    printf("  -mod_ref -- reference modularity for rating function\n");
    printf("  LFR parameters: \n");
    printf("    -mins -- min size of cluster\n");
    printf("    -maxs -- max size of cluster\n");
    printf("    -mind -- min vertex degree\n");
    printf("    -maxd -- max vertex degree\n");
    printf("    -mu   -- mu\n");
    printf("    -dtau -- degree tau\n");
    printf("    -ctau -- community tau\n");
}

void stop_program(int argc, char **argv, graph_t *G)
{
    if (G->rank == 0) { 
        usage(argc, argv);
    }
    MPI_Finalize();
    exit(1);
}


char filename[FNAME_LEN];
LFR_params LFR_p;

void init(int argc, char **argv, graph_t *G)
{
  
    MPI_Init(&argc, &argv);
    MPI_Comm_size(MPI_COMM_WORLD, &G->nproc);
    MPI_Comm_rank(MPI_COMM_WORLD, &G->rank);
 
    LFR_p.mu = 0;
    LFR_p.ctau = 0;
    LFR_p.dtau = 0;
    LFR_p.maxd = 0;
    LFR_p.mind = 0;
    LFR_p.maxs = 0;
    LFR_p.mins = 0;
    
    for (int i = 1; i < argc; ++i) {
        if (!strcmp(argv[i], "-in")) {
            int l = strlen(argv[++i]);
            strncpy(filename, argv[i], (l > FNAME_LEN - 1 ? FNAME_LEN - 1 : l) );
            continue;
        }
        if (!strcmp(argv[i], "-s")) {
            if (argc <= i+1) cout << "Error: should enter the scale" << endl;
            G->scale = (int)atoi(argv[++i]);
            G->n = (vertex_id_t)1 << G->scale;
            continue;
        }
        if (!strcmp(argv[i], "-mu")) {
            LFR_p.mu = (int)atof(argv[++i]);
            continue;
        }
        if (!strcmp(argv[i], "-ctau")) {
            LFR_p.ctau = (int)atof(argv[++i]);
            continue;
        }
        if (!strcmp(argv[i], "-dtau")) {
            LFR_p.dtau = (int)atof(argv[++i]);
            continue;
        }
        if (!strcmp(argv[i], "-maxs")) {
            LFR_p.maxs = (int)atoi(argv[++i]);
            continue;
        }
        if (!strcmp(argv[i], "-mins")) {
            LFR_p.mins = (int)atoi(argv[++i]);
            continue;
        }
        if (!strcmp(argv[i], "-maxd")) {
            LFR_p.maxd = (int)atoi(argv[++i]);
            continue;
        }
        if (!strcmp(argv[i], "-mind")) {
            LFR_p.mind = (int)atoi(argv[++i]);
            continue;
        }
		if (!strcmp(argv[i], "-mod_ref")) {
			modularity_reference = atof(argv[++i]);
        }
        else
        {
            stop_program(argc,argv,G);
        }
    }

}

int main(int argc, char **argv)
{
    graph_t graph;

    graph.scale = 23;
    
    init(argc, argv, &graph);
    
    gen_LFR_graph_MPI(&graph);
    
    MPI_Barrier(MPI_COMM_WORLD);	
    
    // preprocess 1 (without using weights)
    // you can make a convenient form of a graph
    void* graph_user;
	graph_user = build_graph_structure(graph);
	
	
    // find communities
    vector<vertex_id_t> component_id(graph.local_n);
    MPI_Barrier(MPI_COMM_WORLD);
    double start_time = MPI_Wtime();
    if(!graph.rank) cout << "main algorithm have started..." << endl;
	community_detection(graph.n, graph.local_n, graph_user, component_id);
    MPI_Barrier(MPI_COMM_WORLD);
	
    double time = MPI_Wtime() - start_time;
    if (!graph.rank) {
		cout.precision(6);
    	cout << "scale: " << graph.scale << endl;
    	cout << "time: " << fixed << time << " seconds" << endl;
	}
	
    // verify solution (compute modularity)
    double final_modularity = modularity(graph, component_id);
    if (!graph.rank) {
		cout << "modularity: " << final_modularity << endl;
	}
	
    modularity_reference = (modularity_reference > 0) ? modularity_reference : final_modularity; 

    double ratio = final_modularity / modularity_reference;
    if (!graph.rank) {
		cout << "rating: " << fixed << rating(time, ratio) << endl;
	}
    
    free_graph_structure(graph_user);
    freeGraph(&graph);
    MPI_Finalize();
    return 0;
}
